/*****************************************************************************
 *                                                                           *
 *   mail2html.c   - source code for the mail2html program                   *
 *                                                                           *
 *                   written by Tom Georges 8/12/1993                        *
 *                                                                           *
 *   Copyright 1993  Harris Corporation.  ALL RIGHTS RESERVED                *
 *                                                                           *
 *                                                                           *
 *   Change history -                                                        *
 *                                                                           *
 *                     8/14/93 - added author and date to list and correct   *
 *                               title (includes name of folder).  Updated   *
 *                               makefile for conditional on author/date.    *
 *                               Consolodated Find functions into one.       *
 *                                                                           *
 *                     8/18/93 - consolidated headings into structure.       *
 *                               Added hotlinks for the "In-Reply-To" field  *
 *                               to point to originating message id.         *
 *                               Added debug as param 4 on command line      *
 *                                                                           *
 *****************************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>

#include "mail2html.h"           /* source header file */

/* Globals */

int  num_bytes;                         /* total chars in mailbox file */
int  message_len;                       /* length of an individual message */
char output_filename[MAX_FILENAME];     /* top page filename */
char mail_dir[MAX_FILENAME];            /* mail "folder" subdirectory */
int  mail_count = 1;
int  debug = FALSE;


int
PrintUsage ()
  {
  printf("%s\n", USAGE_STRING);
  }

int
main (argc, argv)
int argc;
char **argv;
  {
  int  i;                                /* to make sure we don't pass EOF */
  int  fd;                               /* file descriptor for input file */
  char *pointer;                         /* pointer within total mailbox */
  char *big_file_buf;                    /* all mail messages */
  char *file_buf;                        /* a single mail message */
  Mail_header  *mail_header;             /* message fields */
  Mail_header  *this_message;

  if (argc < 4)
    {
    PrintUsage();
    exit();
    }
  if (strcmp(argv[4], "debug") == 0)
    {
    debug = TRUE;
    }
  
  if (debug)
    {    
    printf("Using file '%s'\n", argv[1]);
    printf("output file '%s'\n", argv[2]);
    printf("mail_dir file '%s'\n", argv[3]);
    }

  strcpy(output_filename, argv[2]);
  strcpy(mail_dir, argv[3]);

  fd = open(argv[1], O_RDONLY);         /* open mailbox file - read only */
  if (fd == -1)
    {
    fprintf(stderr, "Could not open %s\n", argv[1]);
    exit(1);
    }
  num_bytes = lseek(fd, 0, SEEK_END);             /* find size of file */

  big_file_buf = (char *)malloc(num_bytes + 1);   /* allocate space */
  file_buf = (char *)malloc(num_bytes + 1);       /* allocate space */
  
  lseek(fd, 0, SEEK_SET);                         /* reset to beginning */
  read(fd, big_file_buf, num_bytes);              /* read it into buffer */ 
  close(fd);                                      /* close file */
  
  big_file_buf[num_bytes] = 0;                    /* make a NT string */
  
  CreateHtmlMailPageTop();                        /* create header */

  mail_header = (Mail_header *)malloc(sizeof(Mail_header));
  this_message = mail_header;
  this_message->next_message = NULL;

  /* break the file up into multiple messages */
  
  i = 0; 
  pointer = big_file_buf;                         /* start at top of buf */
  while (i < num_bytes && pointer != NULL)        /* don't go past EOF */
    {

    message_len = FindNextMessage(&pointer, file_buf, big_file_buf);
    if (message_len == 0)
      {
      this_message = NULL;
      break;                          /* last message - get out and finish */
      }
    /* init all fields to null */

    strcpy(this_message->subject, "No Subject"); 
    strcpy(this_message->author, "");              
    strcpy(this_message->date, "");               
    strcpy(this_message->reply, "");               

    FindSection(file_buf, this_message->subject, "Subject:"); 

    if (debug)
      {
      printf("Subject found: %s\n", this_message->subject);
      }
    FindSection(file_buf, this_message->msg_id, "Message-ID:"); 
    FindSection(file_buf, this_message->msg_id, "Message-Id:"); 
    StripHTML(this_message->msg_id);
    if (debug)
      {
      printf("message id  found: '%s'\n", this_message->msg_id);
      }
    strcpy(this_message->html_msg_id, this_message->msg_id);
    FindSection(file_buf, this_message->reply, "In-Reply-To:");
    StripHTML(this_message->reply);
#ifdef AUTHOR
    FindSection(file_buf, this_message->author, "From:");     
#endif
#ifdef DATE
    FindSection(file_buf, this_message->date, "Date:");      
#endif

    /* copy message text */
    this_message->text = (char *)malloc(strlen(file_buf) + 1);
    strcpy(this_message->text, file_buf);

    CreateHtmlEntryForSubject(this_message);              /* format anchor */
    CreateHtmlMailPageEntry(this_message->html_subject);  /* create hotlink */

    /* point to next message */
    mail_count++;
    this_message->next_message = (Mail_header *)malloc(sizeof(Mail_header));
    this_message = this_message->next_message;
    this_message->next_message = NULL;

    }
  CreateHtmlMailPageBottom();
  free(big_file_buf);
  free(file_buf);

  /* now create the hotlinks to the "In-Reply-To" fields */

  CreateReplyHotlinks(mail_header);    /* point to top of list */
  InsertAnchorHotlinks(mail_header);
  }

int
FindNextMessage (pointer, file_buf, big_file_buf)
char **pointer;
char *file_buf;
char *big_file_buf;
  {
  char *begin;
  char *end;
  char *last;
  int  msg_len;
   
  begin = strstr(*pointer, "From ");       /* assume a From at top */
  if (begin == NULL)
    {
    return(0);
    }
  end = strstr(begin + 6, "From ");        /* top of next message */
  if (end == NULL)                         /* last message */
    {
    msg_len = big_file_buf + num_bytes - begin;  /* back up if last message */
    }
  else
    {
    msg_len = end - begin;                 /* calculate message length */
    }
  strncpy(file_buf, begin, msg_len);       /* use pointers to copy message */
  file_buf[msg_len] = 0;                   /* null terminate */
  last = end;                              
  *pointer = last;                         /* move pointer to next message */
  return(msg_len);
  }

int
FindSection (buf, section, token)
char        *buf;
char        *section;
char        *token;
  {
  int  i;
  char *found; 
  char *pointer;
  
  pointer = buf;                              /* point to beginning */
  found = strstr(buf, token);                 /* find token */
  if (found == NULL)
    {
    return(0);
    } 
  pointer = found + strlen(token) + 1;        /* point to text */
  found = pointer;

  i = 0;
  while ((*found) != LINE_FEED)               /* look for LF (end of auth) */
    {
    found++;
    i++;
    }
  strncpy(section, pointer, i);               /* copy section */
  section[i] = 0;                              /* null terminate */
  return(1);
  }

int
StripHTML (buffer)
char *buffer;
  {
  int i;
  int j;
  int len;
  
  len = strlen(buffer);
  if (len == 0)
    {
    return;
    }
  StripLeadingWhite(buffer);

  for (i = 0; i < len; i++)
    {
    if (buffer[i] == '<')
      {
      for (j = i; j < len; j++)
        {
        buffer[j] = buffer[j + 1];
        }
      }
    }
  for (i = 0; i < len; i++)
    {
    if (buffer[i] == '>') 
      {
      for (j = i; j < len - 1; j++)
        {
        buffer[j] = buffer[j + 1];
        }
      }
    }
  }

int
StripLeadingWhite (buffer)
char *buffer;
  {
  int i;
  int len;
  int found = 0;
  char *newbuf;
  
  len = strlen(buffer);

  newbuf = (char *)malloc(len + 1);

  for (i = 0; i < len; i++)
    {
    if (buffer[i] == ' ' && buffer[i + 1] != ' ')
      {
      found = i + 1;
      break;
      }
    }
  for (i = 0; i < len; i++)
    {
    newbuf[i] = buffer[found + i];
    }
  newbuf[found + i] = 0;
  strcpy(buffer, newbuf);
  }


/* Formats the unordered list entry on the top page */

int
CreateHtmlEntryForSubject (mail_header)
Mail_header *mail_header;
  {

  sprintf(mail_header->html_subject, 
         "<LI><A HREF=\"%s/mail_%d\">%s</A> %s %s</A>\n", 
          mail_dir,  
          mail_count, 
          mail_header->subject, 
          mail_header->author, 
          mail_header->date);
  }

/* Creates the Header on the top page */

int
CreateHtmlMailPageTop ()
  {
  int fd;
  int num;
  int header_len;
  char header[MAX_HEADER];
  char mail_image[MAX_IMAGE];
  
  strcpy(mail_image, MAIL_IMAGE);

  sprintf(header, "<TITLE>Mail for %s</TITLE>\n<H2><img src=\"%s\"> \nMail for %s</H2>\nThis mail page was generated by the program mail2html.\n For more information on getting it, click \n<a href=\"http://neptune.corp.harris.com/mail2html.html\">here</a><p><UL>\n", mail_dir, mail_image, mail_dir);

  header_len = strlen(header);

  fd = open(output_filename, O_RDWR | O_CREAT | O_TRUNC, 0666);

  if (fd == -1)
    {
    printf("Could not create mail page %s\n", output_filename);
    exit(-1);
    }
  num = write(fd, header, header_len);
  close(fd);
  }

/* Creates the footer on the top page */

int
CreateHtmlMailPageBottom ()
  {
  int fd;
  int num;
  int footer_len;

  fd = open(output_filename, O_APPEND | O_RDWR);

  if (fd == -1)
    {
    printf("Could not open for append/rdwr mail page %s\n", output_filename);
    exit(-1);
    }

  footer_len = strlen(FOOTER);
  num = write(fd, FOOTER, footer_len);
  close(fd);
  }

/* Creates an <li> entry for the formatted subject */

int
CreateHtmlMailPageEntry (html_subject)
char *html_subject;
  {
  int fd;
  int num;
  int sub_len;

  sub_len = strlen(html_subject);

  fd = open(output_filename, O_RDWR | O_APPEND);

  if (fd == -1)
    {
    printf("Could not create mail page %s for append\n", output_filename);
    exit(-1);
    }
  num = write(fd, html_subject, sub_len);
  close(fd);
  }


int
CreateHtmlMailEntry (file_buf, count)
char *file_buf;
int  count;
  {
  int fd;
  int num;
  char filename[MAX_FILENAME];
  
  sprintf(filename, "%s/%s/mail_%d", BASE_NAME, mail_dir, count);
  fd = open(filename, O_RDWR | O_CREAT | O_TRUNC, 0666);
  if (fd == -1)
    {
    printf("Could not create %s\n", filename);
    return(-1);
    }
  
  write(fd, "<pre>\n", 6);
  num = write(fd, file_buf, strlen(file_buf));
  close(fd);
  }

int
CreateReplyHotlinks (mail_header)
Mail_header *mail_header;
  {
  Mail_header *h_ptr;
  Message_ids *msg_id;
  Message_ids *m_ptr;
  Message_ids *msg_head;
  int i;
  int j;
  
  /* first get an array of all the msg_ids */

  msg_head = (Message_ids *)malloc(sizeof(Message_ids));
  m_ptr = msg_head;

  for(h_ptr = mail_header,i=1; i < mail_count; h_ptr = h_ptr->next_message, i++)
    {
    if (debug)
      {
      printf("Subject: '%s'\n", h_ptr->subject);
      }
    msg_id = (Message_ids *)malloc(sizeof(Message_ids));
    strcpy(msg_id->message_id, h_ptr->msg_id);
    if (debug)
      {
      printf("Message id.a = '%s'\n", msg_id->message_id);
      }
    m_ptr->next_id = msg_id;
    msg_id->next_id = NULL;
    m_ptr = msg_id;
    }
  
  for (h_ptr = mail_header,i = 1; i < mail_count; 
       h_ptr = h_ptr->next_message, i++)
    {
    for (m_ptr = msg_head->next_id, j = 1; m_ptr; m_ptr = m_ptr->next_id, j++)
      {
      if (strcmp(h_ptr->reply, m_ptr->message_id) == 0 &&
          strcmp(m_ptr->message_id, "") != 0)  /* match!! */
        {
        if (debug)
          {
          printf("Message id = '%s'\n", m_ptr->message_id);
          printf("Found a match...msg %d references msg %d\n", i, j);
          }
        ConvertReplyToAnchor(h_ptr->reply, h_ptr->html_reply, j);  
        if (debug)
          {
          printf("After convert to anch:%s\n", h_ptr->html_reply);
          }
        break; /* GET OUT! */
        }
      else      /* use normal reply (no anchor) */
        {
        strcpy(h_ptr->html_reply, h_ptr->reply);
        }
      }
    FormatText(h_ptr->msg_id, h_ptr->html_msg_id);
    }
  }

int
ConvertReplyToAnchor (reply_to, html_reply, reference_msg)
char *reply_to;
char *html_reply;
int  reference_msg;
  {
  char buffer[MAX_REPLY];
  int i;
  int len;

  strcpy(buffer, reply_to);        /* copy original reply-to */  
  sprintf(html_reply, "<A HREF=\"mail_%d\">%s</A>", 
          reference_msg, buffer);
  if (debug)
    {
    printf("New reply-to: '%s'\n", html_reply);
    }
  }

int
FormatText (input, output)
char *input;
char *output;
  {
  sprintf(output, "%s", input);
  }

int
InsertAnchorHotlinks (mail_header)
Mail_header *mail_header;
  {
  Mail_header *h_ptr;
  int         i;

  for (h_ptr = mail_header, i = 1; h_ptr; h_ptr = h_ptr->next_message, i++)
    {
    if (h_ptr->reply != "")
      {
      if (debug)
        {
        printf("Ready to REPLACE '%s' with '%s'\n", 
                h_ptr->reply, h_ptr->html_reply);
        }
      ReplaceSection(h_ptr->text, h_ptr->reply, h_ptr->html_reply);
      }
    ReplaceSection(h_ptr->text, h_ptr->msg_id, h_ptr->html_msg_id);
    CreateHtmlMailEntry(h_ptr->text, i);  /* dump the message to the file */
    }
  }

int
ReplaceSection (buf, in_text, out_text)
char *buf;                                  /* total buffer */
char *in_text;                              /* text to replace */
char *out_text;                             /* replacement text */
  {
  int  i;
  int  out_len;
  int  buf_len;
  int  old_len;
  int  diff_in_len;
  char *found; 
  char *start;
  char *end;
  char *before_buf;         /* text before the reply-to field */
  char *after_buf;          /* text after the reply-to field */
  
  found = strstr(buf, in_text);               /* find text to replace */
  if (found == NULL)
    {
    if (debug)
      {
      printf("Could not find '%s'\n", in_text);
      }
    return(0);
    } 
  start = found;                         /* beginning of field */

  i = 0;
  while ((*found) != LINE_FEED)               /* look for LF (end of text) */
    {
    found++;
    i++;
    }
  end = found;                            /* point to end of field */
  
  old_len = end - start;                      /* length of old field */
  buf_len = strlen(buf);
  out_len = strlen(out_text);

  diff_in_len = out_len - old_len;

  before_buf = (char *)malloc(buf_len);
  strncpy(before_buf, buf, start - buf);    /* text before reply field */
  before_buf[start - buf - 1] = 0;
 
  after_buf = (char *)malloc(buf_len);
  strcpy(after_buf, end);                    /* text after reply field */
  
  if (debug)
    {
    printf("BEFORE:\n%s\nAFTER\n%s\n", before_buf, after_buf);
    printf("Inserting '%s' between BEFORE and AFTER\n", out_text);
    }

  sprintf(buf, "%s%s%s", before_buf, out_text, after_buf);
  
  buf_len = buf_len + diff_in_len;
  buf[buf_len] = 0;
  if (debug)
    {
    printf("NEWBUF.......\n%s\n", buf);
    }
  return;
  }
